"""
The esa-2scm package is an implementation of the ESA-2SCM algorithm (Sanghoon Lee, 2024)
For documentation and algorithm/methodology details, please refer to my original article: http://www.snbperi.org/article/230

Should you use this package, please cite my article as follows:
- Lee, Sanghoon (2024). ESA-2SCM for Causal Discovery: Causal Modeling with Elastic Segmentation-based Synthetic Instrumental Variable, SnB Political and Economic Research Institute, 1, 21. <snbperi.org/article/230>


   Copyright 2024 Sanghoon Lee (DSsoli). All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


import numpy as np
import scipy.stats as ss
import warnings


class SynIV:
    
    _warning_status = False
    
    @staticmethod
    def dense_rank(x):
        z = ss.rankdata(x, method="dense")
        if len(np.unique(z))==1:
            warnings.warn("0 variance for given sample", UserWarning)
            SynIV._warning_status = True
        
        return z
    
    @staticmethod
    def minmax_scale(x):
        x_scaled = []
        for datum in x:
            datum_scaled = (datum - min(x)) / (max(x) - min(x))
            x_scaled.append(datum_scaled)
        return np.array(x_scaled)
    
    @staticmethod
    def check_concentration(x, M, data_type, tau):
        if data_type not in ['discrete', 'continuous']:
            raise ValueError("data type should be either 'discrete' or 'continuous'")
        n = len(x)
        if data_type == 'discrete':
            unique_elements, counts = np.unique(x, return_counts=True)
            max_concentration_threshold = 1 / M
            max_concentration = np.max(counts) / n
            return counts, max_concentration, max_concentration_threshold, max_concentration <= (max_concentration_threshold + tau)
        elif data_type == 'continuous':
            x_scaled = SynIV.minmax_scale(x)
            T = 1 / M
            C_v_list = []
            for i in range(M):
                C_v = len([x for x in x_scaled if x >= i * T and x < (i+1) * T])
                if i == M-1:
                    C_v = len([x for x in x_scaled if x >= i * T and x <= 1])
                C_v_list.append(C_v)
            max_concentration = np.max(C_v_list) / n
            return C_v_list, max_concentration, T, max_concentration <= (T + tau)
        
    @staticmethod
    def esa(x, M=2, data_type='continuous', tau=0):
        n = len(x)
        idx_sorted = np.argsort(x)
        original_M = M
                
        if M < 2: raise ValueError("Minimum number of segments (M) must be a positive integer greater than or equal to 2")

        fixed_thres = 1/M
        while M >= 2:
            if n % M != 0:
                M -= 1
            _, max_concent, max_concent_threshold, threshold_check = SynIV.check_concentration(x=x, M=M, data_type=data_type, tau=tau)
            if threshold_check:
                break
            M -= 1
        
        if M < 2:
            warnings.warn(f"""
                            Data is excessively concentrated on a single segment to perform meaningful ESA. Using Dense Rank method instead. This may indicate bias in the dataset.
                            Consider either checking data imbalance or increasing the regularization term tau.
                            (Single segment under post-optimization accounts for {max_concent*100:.2f}% of the total dataset while single segment threshold for M={original_M} is fixed at {fixed_thres*100:.2f}%).
                            """, UserWarning)
            return SynIV.dense_rank(x)
        
        if M != original_M:
            warnings.warn(f"""
                            Data is excessively concentrated on a single segment to perform meaningful ESA. Using M={M} instead of M={original_M}.
                            Consider either checking data imbalance or increasing the regularization term tau.
                            (Single segment under post-optimization accounts for {max_concent*100:.2f}% of the total dataset while single segment threshold for M={original_M} is fixed at {fixed_thres*100:.2f}%).
                            """, UserWarning)

        segment_sizes = [n // M + (1 if i < n % M else 0) for i in range(M)]
        boundaries = np.cumsum(segment_sizes)
        z = np.zeros(n, dtype=int)  
        
        start_idx = 0
        for assign_segment_value, boundary in enumerate(boundaries, start=1):
            segment_indices = idx_sorted[start_idx:boundary]
            z[segment_indices] = assign_segment_value
            start_idx = boundary
        
        return z
    
    
    @staticmethod
    def m_split(x, strategy: str = 'auto'):
        strategies = ['auto', 'median', 'mean']
        if strategy not in strategies: raise ValueError(f"Invalid strategy name '{strategy}'")
        
        mean_val, med_val = np.mean(x), np.median(x)
        
        def auto_strategy():
            if med_val != np.max(x):
                z = np.array([1 if i > med_val else -1 for i in x])
                if len(np.unique(z))==1:
                    z = np.array([1 if i > mean_val else -1 for i in x])
                    if len(np.unique(z))==1:
                        z = SynIV.dense_rank(x)
                        
            else: z = SynIV.dense_rank(x)
            return z
        
        def median_strategy():
            return np.array([1 if i > med_val else -1 for i in x])
        
        def mean_strategy():
            return np.array([1 if i > mean_val else -1 for i in x])
        
        strategy_funcs = {
            'auto': auto_strategy,
            'median': median_strategy,
            'mean': mean_strategy
        }
        
        z = strategy_funcs[strategy]()
        
        if len(np.unique(z))==1 and not SynIV._warning_status:
            warnings.warn("0 variance for given sample", UserWarning)
            SynIV._warning_status = True
            
        return z
    
    
    @staticmethod
    def get_syniv(method="esa"):
        syniv_map = {
            "esa": SynIV.esa,
            "dense_rank": SynIV.dense_rank,
            "m_split": SynIV.m_split
        }
        
        try:
            return syniv_map[method]
        except:
            raise ValueError(f"Invalid method name '{method}'")


def r2_score(y_true, y_pred):
    r2 = 1 - np.sum((y_true - y_pred)**2) / np.sum((y_true - np.mean(y_true))**2)
    return r2



"""
The esa-2scm package is an implementation of the ESA-2SCM algorithm (Sanghoon Lee, 2024)
For documentation and algorithm/methodology details, please refer to my original article: http://www.snbperi.org/article/230

Should you use this package, please cite my article as follows:
- Lee, Sanghoon (2024). ESA-2SCM for Causal Discovery: Causal Modeling with Elastic Segmentation-based Synthetic Instrumental Variable, SnB Political and Economic Research Institute, 1, 21. <snbperi.org/article/230>


   Copyright 2024 Sanghoon Lee (DSsoli). All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


import numpy as np
import pandas as pd
from typing import Union, List
from .syniv import SynIV, r2_score


class BaseCM:
    
    _KNOWLEDGE = (None, 'x2->x1', 'x1->x2')
    
    def __init__(
        self, 
        x1: Union[np.ndarray, pd.Series, List], 
        x2: Union[np.ndarray, pd.Series, List],
        prior_knowledge: str = None,
        ):
        
        if np.ndim(x1) > 1 or np.ndim(x2) > 1:
            raise ValueError("Inputs for x1 and x2 must be 1-Dimensional")
        
        if len(x1) != len(x2):
            raise ValueError("Need equal sample sizes for x1 and x2")
        
        try: check_x1, check_x2 = np.array(x1, dtype=float), np.array(x2, dtype=float)
        except ValueError as e:
            raise ValueError("Inputs contain non-numerical values") from None
        
        if np.isnan(check_x1).any() or np.isnan(check_x2).any():
            raise ValueError("Inputs contain NaN values")
        
        if prior_knowledge in BaseCM._KNOWLEDGE:
            self._prior_knowledge = prior_knowledge
        else: raise ValueError("Valid expression for prior_knowledge: 'x2->x1' or 'x1->x2'")
        
        self._x1, self._x2 = np.array(x1) , np.array(x2)


class Esa2Scm(BaseCM):
    
    def __init__(self, x1, x2, prior_knowledge=None):
        super().__init__(x1, x2, prior_knowledge)


    @staticmethod
    def _estimate(x, z):
        if np.var(z, ddof=1) == 0:
            b = 0
        else:  
            b = np.cov(x, z)[0,1] / np.var(z, ddof=1)
        c = np.mean(x) - b * np.mean(z)
        x_hat = c + b * z
        return x_hat, b
    
    
    def fit(self, syniv_method: str = "esa", **kwargs):
        
        if self._prior_knowledge is not None:
            
            def estimate(x1, x2, prior_knowledge):
                x_hat, b = Esa2Scm._estimate(x1, x2)
                causal_dir = prior_knowledge
                score = r2_score(x1, x_hat)
                return x_hat, b, causal_dir, score
                        
            if self._prior_knowledge == 'x2->x1':
                self._x_hat, self._b, self._causal_dir_fixed, self._determined_mod_ols_score = estimate(self._x1, self._x2, self._prior_knowledge)
                
            else:
                self._x_hat, self._b, self._causal_dir_fixed, self._determined_mod_ols_score = estimate(self._x2, self._x1, self._prior_knowledge)

        else:
            estimate_syniv = SynIV.get_syniv(syniv_method)
            self._z1, self._z2 = estimate_syniv(self._x1, **kwargs), estimate_syniv(self._x2, **kwargs)
            
            self._x1_iv1sls, self._x2_iv1sls = Esa2Scm._estimate(self._x1, self._z1)[0], Esa2Scm._estimate(self._x2, self._z2)[0]
            self._x1_slsiv_corr_ = 'undefined (0 variance)' if np.var(self._x1_iv1sls) == 0 or np.var(self._x1) == 0 else np.corrcoef(self._x1_iv1sls, self._x1)[0][1] 
            self._x2_slsiv_corr_ = 'undefined (0 variance)' if np.var(self._x2_iv1sls) == 0 or np.var(self._x2) == 0 else np.corrcoef(self._x2_iv1sls, self._x2)[0][1]
            
            self._x1_iv2sls, self._b12 = Esa2Scm._estimate(self._x1, self._x2_iv1sls)
            self._x2_iv2sls, self._b21 = Esa2Scm._estimate(self._x2, self._x1_iv1sls)
            
            self._score, self._score_rev = round(r2_score(self._x1, self._x1_iv2sls), 5), round(r2_score(self._x2, self._x2_iv2sls), 5) 
            
            self._causal_dir = "x2->x1" if self._score > self._score_rev else "x1->x2" if self._score < self._score_rev else "undetermined"
            
            if self._causal_dir == "x2->x1":
                self._true_x_hat, self._true_causal_coef, self._true_score = self._x2_iv2sls, self._b12, self._score
                self._determined_mod_ols_score = round(r2_score(self._x1, Esa2Scm._estimate(self._x1, self._x2)[0]), 5) 
                
            elif self._causal_dir == "x1->x2":
                self._true_x_hat, self._true_causal_coef, self._true_score = self._x1_iv2sls, self._b21, self._score_rev
                self._determined_mod_ols_score = round(r2_score(self._x2, Esa2Scm._estimate(self._x2, self._x1)[0]), 5) 
                
            else:
                self._true_x_hat, self._true_causal_coef, self._true_score, self._determined_mod_ols_score = np.nan, 0, 0, 0
            
        return self
    
    
    def _summary(self):
        summary_idx = ["Causal Direction", "Causal Coefficient", "Goodness of Fit"]
        additional_idx = ["Corr (2SLS_IV-Explanatory)", "Posthoc Goodness of Fit"]
        
        if self._prior_knowledge is not None:
            self._result = pd.DataFrame({self._prior_knowledge + " (Predetermined)": [self._prior_knowledge, self._b, self._determined_mod_ols_score]}, index=summary_idx)
            return self._result
        
        summary_columns = ['x2->x1', 'x1->x2']
        summary_values = [str(self._causal_dir == 'x2->x1'), self._b12, self._score, self._x2_slsiv_corr_, [self._determined_mod_ols_score if self._causal_dir == 'x2->x1' else '-'][0]]
        summary_values_rev = [str(self._causal_dir == 'x1->x2'), self._b21, self._score_rev, self._x1_slsiv_corr_, [self._determined_mod_ols_score if self._causal_dir == 'x1->x2' else '-'][0]]
        self._result = pd.DataFrame({summary_columns[0]: summary_values, summary_columns[1]: summary_values_rev}, index=summary_idx+additional_idx)
        
        return self._result
    
    
    @property
    def x1(self):
        return self._x1
    
    @property
    def x2(self):
        return self._x2
    
    @property
    def z1(self):
        if hasattr(self, "_z1"):
            return self._z1
        raise AttributeError("Synthetic IV (z1) is not generated as Prior Knowledge has been set")
    
    @property
    def z2(self):
        if hasattr(self, "_z2"):
            return self._z2
        raise AttributeError("Synthetic IV (z2) is not generated as Prior Knowledge has been set")
    
    @property
    def causal_coef(self):
        if hasattr(self, "_b"):
            return self._b
        return self._true_causal_coef
    
    @property
    def causal_direction(self):
        if hasattr(self, "_causal_dir_fixed"):
            return self._causal_dir_fixed + ' (prior knowledge)'
        return self._causal_dir
    
    @property
    def esa2scm_score(self):
        if hasattr(self, "_true_score"):
            return self._true_score
        raise AttributeError("esa2scm_score is not unavailable as Pior Knowledge has been set")
    
    @property
    def posthoc_score(self):
        return self._determined_mod_ols_score
    
    @property
    def x_hat(self):
        if hasattr(self, "_x_hat"):
            return self._x_hat
        return self._true_x_hat
    
    @property
    def corr_x1_to_slsiv(self):
        if hasattr(self, "_x1_slsiv_corr_"):
            return self._x1_slsiv_corr_
        raise AttributeError("Synthetic IV to calculate correlation coefficient is not generated as Prior knowledge has been set")
    
    @property
    def corr_x2_to_slsiv(self):
        if hasattr(self, "_x2_slsiv_corr_"):
            return self._x2_slsiv_corr_
        raise AttributeError("Synthetic IV to calculate correlation coefficient is not generated as Prior knowledge has been set")
    
    @property
    def summary(self):
        return self._summary
    



"""
The esa-2scm package is an implementation of the ESA-2SCM algorithm (Sanghoon Lee, 2024)
For documentation and algorithm/methodology details, please refer to my original article: http://www.snbperi.org/article/230

Should you use this package, please cite my article as follows:
- Lee, Sanghoon (2024). ESA-2SCM for Causal Discovery: Causal Modeling with Elastic Segmentation-based Synthetic Instrumental Variable, SnB Political and Economic Research Institute, 1, 21. <snbperi.org/article/230>


   Copyright 2024 Sanghoon Lee (DSsoli). All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


import warnings
import numpy as np
from esa_2scm.syniv import SynIV


X_0Var = [1,1,1]
X_M2 = [1,1,2,3]
X_M3 = [1,1,2,3,4,5]


def capture_warning(func, *args, **kwargs):
    
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        func(*args, **kwargs)
    
    if w:
        warning_message = str(w[-1].message)
        if len(w) > 1:
            warning_message = str(w[-2].message) + " " + warning_message
            
        return warning_message
    
    else:
        return "no warnings captured"


def get_res(test_sample):
    warnings.filterwarnings("ignore")
    esa_res = SynIV.esa(test_sample, M=5)
    rank_res = SynIV.dense_rank(test_sample)
    split_res = SynIV.m_split(test_sample, strategy="auto")
    
    return esa_res, rank_res, split_res


def get_warnings(test_sample):
    esa_warn = capture_warning(SynIV.esa, test_sample, M=5)
    rank_warn = capture_warning(SynIV.dense_rank, test_sample)
    split_warn = capture_warning(SynIV.m_split, test_sample, strategy='auto')
    
    return esa_warn, rank_warn, split_warn


def test_res_0var():
    esa_res, rank_res, split_res = get_res(X_0Var)
    assert np.array_equal(esa_res, np.array([1,1,1]))
    assert np.array_equal(rank_res, np.array([1,1,1]))
    assert np.array_equal(split_res, np.array([1,1,1]))


def test_warnings_0var():
    esa_warn, rank_warn, split_warn = get_warnings(X_0Var)
    assert "100.00%" in esa_warn
    assert "20.00%" in esa_warn
    assert "0 variance for given sample" in esa_warn
    assert "0 variance for given sample" in rank_warn
    assert "0 variance for given sample" in split_warn


def test_res_M2():
    esa_res, rank_res, split_res = get_res(X_M2)
    assert np.array_equal(esa_res, np.array([1,1,2,2]))
    assert np.array_equal(rank_res, np.array([1,1,2,3]))
    assert np.array_equal(split_res, np.array([-1,-1,1,1]))


def test_warnings_M2():
    esa_warn, rank_warn, split_warn = get_warnings(X_M2)
    assert "50.00%" in esa_warn
    assert "20.00%" in esa_warn
    assert "no warnings captured" in rank_warn
    assert "no warnings captured" in split_warn


def test_res_M3():
    esa_res, rank_res, split_res = get_res(X_M3)
    assert np.array_equal(esa_res, np.array([1,1,2,2,3,3]))
    assert np.array_equal(rank_res, np.array([1,1,2,3,4,5]))
    assert np.array_equal(split_res, np.array([-1,-1,-1,1,1,1]))


def test_warnings_M3():
    esa_warn, rank_warn, split_warn = get_warnings(X_M3)
    assert "33.33%" in esa_warn
    assert "20.00%" in esa_warn
    assert "no warnings captured" in rank_warn
    assert "no warnings captured" in split_warn



"""
The esa-2scm package is an implementation of the ESA-2SCM algorithm (Sanghoon Lee, 2024)
For documentation and algorithm/methodology details, please refer to my original article: http://www.snbperi.org/article/230

Should you use this package, please cite my article as follows:
- Lee, Sanghoon (2024). ESA-2SCM for Causal Discovery: Causal Modeling with Elastic Segmentation-based Synthetic Instrumental Variable, SnB Political and Economic Research Institute, 1, 21. <snbperi.org/article/230>


   Copyright 2024 Sanghoon Lee (DSsoli). All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


from esa_2scm import Esa2Scm
import numpy as np

def model_simulation():
    correct_count = 0
    wrong_count = 0
    total_runs = 100
    b_list = []

    for _ in range(total_runs):
        np.random.seed(11)
        N = 10000
        x2 = np.random.random(size=N)
        e1 = np.random.normal(size=N)
        b12 = 1.8
        x1 = b12 * x2 + e1
        b_list.append(b12)
        model = Esa2Scm(x1, x2)
        model.fit('esa', M=5)

        if model.causal_direction == "x2->x1":
            correct_count += 1
        else: wrong_count += 1
        
        mean_estimated_causal_impact_coef = np.mean(b_list)
        
    
    return correct_count, wrong_count, mean_estimated_causal_impact_coef, model


def test_model_simulation():
    correct_count, wrong_count, mean_estimated_causal_impact_coef, model = model_simulation()
    
    assert correct_count == 100
    assert wrong_count == 0
    assert np.round(mean_estimated_causal_impact_coef, 2)  == 1.80
    assert round(model.causal_coef, 1) == 1.8
    assert model.causal_direction == "x2->x1"
    assert round(model.esa2scm_score, 2) == 0.21
    assert round(model.posthoc_score, 2) == 0.22
    assert type(model.x1) == np.ndarray
    assert type(model.x2) == np.ndarray
    assert hasattr(model, "x_hat") == True
    assert len(np.unique(model.z1)) == 5
    assert len(np.unique(model.z2)) == 5
    assert hasattr(model, "corr_x1_to_slsiv") == True
    assert hasattr(model, "corr_x2_to_slsiv") == True
    assert hasattr(model, "summary") == True
    

def test_prior_knowledge():
    N = 10000
    shape, scale = 2., 2.
    np.random.seed(11)
    x2 = np.random.gamma(shape, scale, N)
    e1 = np.random.normal(size=N)
    b12 = 1.8
    x1 = b12 * x2 + e1
    model = Esa2Scm(x1, x2, prior_knowledge="x1->x2")
    model.fit()
    assert model.causal_direction == "x1->x2 (prior knowledge)"
    assert round(model.causal_coef, 1) == 0.5
    assert round(model.posthoc_score, 2) == 0.96
    assert hasattr(model, "summary") == True
    assert hasattr(model, "z1") == False
    assert hasattr(model, "syniv2sls_score") == False
    


"""
The esa-2scm package is an implementation of the ESA-2SCM algorithm (Sanghoon Lee, 2024)
For documentation and algorithm/methodology details, please refer to my original article: http://www.snbperi.org/article/230

Should you use this package, please cite my article as follows:
- Lee, Sanghoon (2024). ESA-2SCM for Causal Discovery: Causal Modeling with Elastic Segmentation-based Synthetic Instrumental Variable, SnB Political and Economic Research Institute, 1, 21. <snbperi.org/article/230>


   Copyright 2024 Sanghoon Lee (DSsoli). All Rights Reserved.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    README = fh.read()

setup(
    author="Sanghoon Lee (DSsoli)",
    author_email="solisoli3197@gmail.com",
    name="esa-2scm",
    version="0.2.7",
    description="ESA-2SCM Python Package for Causal Discovery",
    long_description=README,
    long_description_content_type="text/markdown",
    install_requires=["numpy", "pandas", "scipy"],
    url="https://github.com/DSsoli/esa-2scm.git",
    packages=find_packages(include=['esa_2scm', 'esa_2scm.*']),
    package_data={"esa_2scm": ['LICENSE', 'examples/*']},
    include_package_data=True
)